import dataclasses
from functools import partial
from typing import Optional, Literal
from typing import Union, Tuple

import jax
import numpy as np
import tensorflow_probability.substrates.jax as tfp
from jax import numpy as jnp, vmap, lax
from jax._src.scipy.special import gammaln, logsumexp
from tensorflow_probability.python.internal.backend.jax import cumulative_logsumexp

from jaxctx.priors.prior import AbstractPrior
from jaxctx.priors.types import FloatArray

tfpd = tfp.distributions

__all__ = [
    "Bernoulli",
    "Beta",
    "Categorical",
    "ForcedIdentifiability",
    "Poisson",
    "UnnormalisedDirichlet",
    "Empirical",
    "TruncationWrapper",
    "ExplicitDensityPrior",
]

count_dtype = jnp.int64
measure_dtype = jnp.float64


class Bernoulli(AbstractPrior):
    def __init__(self, *, logits=None, probs=None, name: Optional[str] = None, base_dtype=jnp.float32):
        super(Bernoulli, self).__init__(name=name, base_dtype=base_dtype)
        self.dist = tfpd.Bernoulli(logits=logits, probs=probs)

    def _dtype(self):
        return self.dist.dtype

    def _base_shape(self):
        return self._shape()

    def _shape(self):
        return tuple(self.dist.batch_shape_tensor()) + tuple(self.dist.event_shape_tensor())

    def _forward(self, U):
        U = U.astype(jnp.result_type(U.dtype, self.dist.dtype))
        return self._quantile(U)

    def _inverse(self, X):
        return self.dist.cdf(X)

    def _log_prob(self, X):
        return self.dist.log_prob(X)

    def _quantile(self, U):
        probs = self.dist._probs_parameter_no_checks()
        sample = jnp.less(U, probs)
        return sample.astype(self.dtype)


class Beta(AbstractPrior):
    def __init__(self, *, concentration0=None, concentration1=None, name: Optional[str] = None, base_dtype=jnp.float32):
        super(Beta, self).__init__(name=name, base_dtype=base_dtype)
        # Special cases for Beta that are faster use the Kumaraswamy distribution
        if isinstance(concentration0, (float, int)) and concentration0 == 1:
            self.dist = tfpd.Kumaraswamy(concentration0=concentration0, concentration1=concentration1)
        elif isinstance(concentration1, (float, int)) and concentration1 == 1:
            self.dist = tfpd.Kumaraswamy(concentration0=concentration0, concentration1=concentration1)
        elif isinstance(concentration0, np.ndarray) and np.all(concentration0 == 1):
            self.dist = tfpd.Kumaraswamy(concentration0=concentration0, concentration1=concentration1)
        elif isinstance(concentration1, np.ndarray) and np.all(concentration1 == 1):
            self.dist = tfpd.Kumaraswamy(concentration0=concentration0, concentration1=concentration1)
        else:
            self.dist = tfpd.Beta(concentration0=concentration0, concentration1=concentration1)

    def _dtype(self):
        return self.dist.dtype

    def _base_shape(self):
        return self._shape()

    def _shape(self):
        return tuple(self.dist.batch_shape_tensor()) + tuple(self.dist.event_shape_tensor())

    def _forward(self, U):
        U = U.astype(jnp.result_type(U.dtype, self.dist.dtype))
        return self._quantile(U)

    def _inverse(self, X):
        return self.dist.cdf(X)

    def _log_prob(self, X):
        return self.dist.log_prob(X)

    def _quantile(self, U):
        X = self.dist.quantile(U)
        return X.astype(self.dtype)


class Categorical(AbstractPrior):
    def __init__(self, parametrisation: Literal['gumbel_max', 'cdf'], *, logits=None, probs=None,
                 name: Optional[str] = None, base_dtype=jnp.float32):
        """
        Initialised Categorical special prior.

        Args:
            parametrisation: 'cdf' is good for discrete params with correlation between neighbouring categories,
                otherwise gumbel is better.
            logits: log-prob of each category
            probs: prob of each category
            name: optional name
        """
        super(Categorical, self).__init__(name=name, base_dtype=base_dtype)
        self.dist = tfpd.Categorical(logits=logits, probs=probs)
        self._parametrisation = parametrisation

    def _dtype(self):
        return self.dist.dtype

    def _base_shape(self):
        if self._parametrisation == 'gumbel_max':
            return self._shape() + (self.dist._num_categories(),)
        elif self._parametrisation == 'cdf':
            return self._shape()

    def _shape(self):
        return tuple(self.dist.batch_shape_tensor()) + tuple(self.dist.event_shape_tensor())

    def _forward(self, U):
        U = U.astype(jnp.result_type(U.dtype, self.dist.dtype))
        if self._parametrisation == 'gumbel_max':
            return self._quantile_gumbelmax(U)
        elif self._parametrisation == 'cdf':
            return self._quantile_cdf(U)

    def _inverse(self, X):
        if self._parametrisation == 'cdf':
            return self.dist.cdf(X)
        raise NotImplementedError()

    def _log_prob(self, X):
        return self.dist.log_prob(X)

    def _quantile_gumbelmax(self, U):
        logits = self.dist._logits_parameter_no_checks()
        sample_dtype = self.dtype
        z = -jnp.log(-jnp.log(U))  # gumbel
        draws = jnp.argmax(logits + z, axis=-1).astype(sample_dtype)
        return draws

    def _cdf_gumbelmax(self, X):
        logits = self.dist._logits_parameter_no_checks()
        z = jnp.max(logits, axis=-1, keepdims=True)
        logits -= z
        return jnp.exp(logits - logsumexp(logits, axis=-1, keepdims=True))

    def _quantile_cdf(self, U):
        """
        The quantile for CDF parametrisation.

        Args:
            U: [...]

        Returns:
            [...]
        """
        logits = self.dist._logits_parameter_no_checks()  # [..., N]
        cum_logits = cumulative_logsumexp(logits, axis=-1)  # [..., N]
        cum_logits -= cum_logits[..., -1:]
        N = cum_logits.shape[-1]
        cum_logits_flat = jnp.reshape(cum_logits, (-1, N))
        log_U_flat = jnp.reshape(jnp.log(U), (-1,))

        category_flat = vmap(lambda a, v: jnp.searchsorted(a, v, side='left'))(cum_logits_flat, log_U_flat)
        category = jnp.reshape(category_flat, U.shape)
        return category


class ForcedIdentifiability(AbstractPrior):
    """
    Prior for a sequence of `n` jax.random variables uniformly distributed on U[low, high] such that U[i,...] <= U[i+1,...].
    For broadcasting the resulting jax.random variable is sorted on the first dimension elementwise.

    Args:
        n: number of samples within [low,high]
        low: minimum of distribution
        high: maximum of distribution
        fix_left: if True, the leftmost value is fixed to `low`
        fix_right: if True, the rightmost value is fixed to `high`
    """

    def __init__(self, *, n: int, low=None, high=None, fix_left: bool = False, fix_right: bool = False,
                 name: Optional[str] = None, base_dtype=jnp.float32):
        super(ForcedIdentifiability, self).__init__(name=name, base_dtype=base_dtype)
        n_min = (1 if fix_left else 0) + (1 if fix_right else 0)
        if n < n_min:
            raise ValueError(f'`n` too small for fix_left={fix_left} and fix_right={fix_right}')
        self.n = n
        low, high = jnp.broadcast_arrays(low, high)
        self.low = low
        self.high = high
        self.fix_left = fix_left
        self.fix_right = fix_right

    def _dtype(self):
        return measure_dtype

    def _base_shape(self):
        num_base = self.n
        if self.fix_left:
            num_base -= 1
        if self.fix_right:
            num_base -= 1
        return (num_base,) + np.shape(self.low)

    def _shape(self):
        return (self.n,) + np.shape(self.low)

    def _forward(self, U):
        return self._quantile(U)

    def _inverse(self, X):
        return self._cdf(X)

    def _log_prob(self, X):
        n = self.n
        if self.fix_left:
            n -= 1
        if self.fix_right:
            n -= 1
        log_n_fac = gammaln(n + 1)
        diff = self.high - self.low
        log_prob = - log_n_fac - n * jnp.log(diff)
        # no check that X is inside high and low
        return log_prob

    def _cdf(self, X):
        if self.fix_left:
            X = X[1:]
        if self.fix_right:
            X = X[:-1]

        n = self.n
        if self.fix_left:
            n -= 1
        if self.fix_right:
            n -= 1

        log_output = jnp.log((X - self.low) / (self.high - self.low))

        # Step 2: Undo cumulative sum operation
        inner = jnp.diff(log_output[::-1], prepend=0, axis=0)[::-1]

        # Step 3: Undo reshaping and division
        k = jnp.arange(n) + 1
        log_x = inner * jnp.reshape(k, (n,) + (1,) * (len(self.shape) - 1))

        # Step 4: Find U by exponentiating log_x
        U = jnp.exp(log_x)

        return U.astype(self.dtype)

    def _quantile(self, U):
        n = self.n
        if self.fix_left:
            n -= 1
        if self.fix_right:
            n -= 1
        log_x = jnp.log(U)  # [n, ...]
        k = jnp.arange(n) + 1
        inner = log_x / lax.reshape(k, (n,) + (1,) * (len(self.shape) - 1))
        log_output = jnp.cumsum(inner[::-1], axis=0)[::-1]
        output = self.low + (self.high - self.low) * jnp.exp(log_output)
        if self.fix_left:
            output = jnp.concatenate([self.low[None], output], axis=0)
        if self.fix_right:
            output = jnp.concatenate([output, self.high[None]], axis=0)
        return output.astype(self.dtype)


def _poisson_quantile_bisection(U, rate, max_iter=15, unroll: bool = True):
    # max_iter is set so that error < 1 up to rate of 1e4
    rate = jnp.maximum(jnp.asarray(rate), 1e-5)
    if np.size(rate) > 1:
        raise ValueError("Rate must be a scalar")

    if np.size(U) > 1:
        U_flat = U.ravel()
        x_final, x_results = vmap(lambda u: _poisson_quantile_bisection(u, rate, max_iter, unroll))(U_flat)
        return x_final.reshape(U.shape), x_results.reshape(U.shape + (max_iter,))

    def smooth_cdf(x, rate):
        return lax.igammac(x + 1., rate)

    def fixed_point_update(x, args):
        (a, b, f_a, f_b) = x

        c = 0.5 * (a + b)
        f_c = smooth_cdf(c, rate)

        left = f_c > U
        a1 = jnp.where(left, a, c)
        f_a1 = jnp.where(left, f_a, f_c)
        b1 = jnp.where(left, c, b)
        f_b1 = jnp.where(left, f_c, f_b)

        a2 = a
        f_a2 = f_a
        b2 = b * 2.
        f_b2 = smooth_cdf(b2, rate)

        bounded = jnp.greater_equal(f_b, U)  # a already bounds.

        a = jnp.where(bounded, a1, a2)
        b = jnp.where(bounded, b1, b2)
        f_a = jnp.where(bounded, f_a1, f_a2)
        f_b = jnp.where(bounded, f_b1, f_b2)

        new_x = (a, b, f_a, f_b)

        return new_x, 0.5 * (a + b)

    a = jnp.asarray(0.)
    b = jnp.asarray(rate)
    f_a = jnp.asarray(0.)
    f_b = smooth_cdf(b, rate)
    init = (a, b, f_a, f_b)

    # Dummy array to facilitate using scan for a fixed number of iterations
    (a, b, f_a, f_b), x_results = lax.scan(
        fixed_point_update,
        init,
        jnp.arange(max_iter),
        unroll=max_iter if unroll else 1
    )

    c = 0.5 * (a + b)

    return c, x_results


@partial(jax.jit, static_argnames=("unroll",))
def _poisson_quantile(U, rate, unroll: bool = False):
    x, _ = _poisson_quantile_bisection(U, rate, unroll=unroll)
    return x.astype(count_dtype)


class Poisson(AbstractPrior):
    def __init__(self, *, rate=None, log_rate=None, name: Optional[str] = None, base_dtype=jnp.float32):
        super(Poisson, self).__init__(name=name, base_dtype=base_dtype)
        self.dist = tfpd.Poisson(rate=rate, log_rate=log_rate)

    def _dtype(self):
        return count_dtype

    def _base_shape(self):
        return self._shape()

    def _shape(self):
        return tuple(self.dist.batch_shape_tensor()) + tuple(self.dist.event_shape_tensor())

    def _forward(self, U):
        U = U.astype(jnp.result_type(U.dtype, self.dist.dtype))
        return self._quantile(U)

    def _inverse(self, X):
        return self.dist.cdf(X)

    def _log_prob(self, X):
        return self.dist.log_prob(X)

    def _quantile(self, U):
        """
        Algorithmic Poisson generator based upon the inversion by sequential search

        .. code-block::

                init:
                    Let log_x ← -inf, log_p ← −λ, log_s ← log_p.
                    Generate uniform jax.random number u in [0,1].
                while log_u > log_s do:
                    log_x ← logaddexp(log_x, 0).
                    log_p ← log_p + log_λ - log_x.
                    log_s ← logaddexp(log_s, log_p).
                return exp(log_x).
        """
        rate = self.dist.rate_parameter()
        if np.size(rate) > 1:
            return jax.vmap(lambda u, r: _poisson_quantile(u, r))(U.ravel(), rate.ravel()).reshape(self.shape)
        return _poisson_quantile(U, rate)


class UnnormalisedDirichlet(AbstractPrior):
    """
    Represents an unnormalised dirichlet distribution of K classes.
    That is, the output is related to the K-simplex via normalisation.

    X ~ UnnormalisedDirichlet(alpha)
    Y = X / sum(X) ==> Y ~ Dirichlet(alpha)
    """

    def __init__(self, *, concentration, name: Optional[str] = None, base_dtype=jnp.float32):
        super(UnnormalisedDirichlet, self).__init__(name=name, base_dtype=base_dtype)
        self._dirichlet_dist = tfpd.Dirichlet(concentration=concentration)
        self._gamma_dist = tfpd.Gamma(concentration=concentration, log_rate=0.)

    def _dtype(self):
        return self._dirichlet_dist.dtype

    def _base_shape(self):
        return self._shape()

    def _shape(self):
        return tuple(self._dirichlet_dist.batch_shape_tensor()) + tuple(self._dirichlet_dist.event_shape_tensor())

    def _forward(self, U):
        U = U.astype(jnp.result_type(U.dtype, self._gamma_dist.dtype))
        return self._quantile(U)

    def _inverse(self, X):
        return self._gamma_dist.cdf(X)

    def _log_prob(self, X):
        return jnp.sum(self._gamma_dist.log_prob(X), axis=-1)

    def _quantile(self, U):
        gamma = self._gamma_dist.quantile(U).astype(self.dtype)
        return gamma


class Empirical(AbstractPrior):
    """
    Represents the empirical distribution of a set of 1D samples, with arbitrary batch dimension.
    """

    def __init__(self, *, samples: jax.Array, support_min: Optional[FloatArray] = None,
                 support_max: Optional[FloatArray] = None, resolution: int = 100, name: Optional[str] = None,
                 base_dtype=jnp.float32):
        super(Empirical, self).__init__(name=name, base_dtype=base_dtype)
        if len(np.shape(samples)) < 1:
            raise ValueError("Samples must have at least one dimension")
        if np.size(samples) == 0:
            raise ValueError("Samples must have at least one element")
        if resolution < 1:
            raise ValueError("Resolution must be at least 1")
        samples = jnp.asarray(samples)
        # Add 1 point for each support endpoint
        endpoints = []
        if support_min is not None:
            endpoints.append(support_min)
        if support_max is not None:
            endpoints.append(support_max)
        if len(endpoints) > 0:
            samples = jnp.concatenate([samples, jnp.asarray(endpoints)])

        resolution = min(resolution, len(samples) - 1)

        self._q = jnp.linspace(0., 100., resolution + 1)
        self._percentiles = jnp.reshape(jnp.percentile(samples, self._q, axis=-1), (resolution + 1, -1))

        self._batch_shape = np.shape(samples)[:-1]
        self._samples_dtype = samples.dtype

    def _dtype(self):
        return self._samples_dtype

    def _base_shape(self):
        return self._batch_shape

    def _shape(self):
        return self._batch_shape

    def _forward(self, U):
        return self._quantile(U)

    def _inverse(self, X):
        return self._cdf(X)

    def _cdf(self, X):
        X_flat = jnp.ravel(X)
        u = jax.vmap(lambda x, per: jnp.interp(x, per, self._q * 1e-2), in_axes=(0, 1))(X_flat, self._percentiles)
        u = lax.reshape(u, X.shape)
        return u

    def _log_prob(self, X):
        X_flat = jnp.ravel(X)

        def _cdf(x):
            return self._cdf(x)

        def log_pdf(x):
            return jnp.log(jax.grad(_cdf)(x))

        log_prob = lax.reshape(jax.vmap(log_pdf)(X_flat), X.shape)
        return log_prob

    def _quantile(self, U):
        U_flat = jnp.ravel(U)
        x = jax.vmap(lambda u, per: jnp.interp(u * 100., self._q, per), in_axes=(0, 1))(U_flat, self._percentiles)
        x = lax.reshape(x, U.shape)
        return x


class TruncationWrapper(AbstractPrior):
    """
    Wraps another prior to make it truncated.

    For truncated distribution the quantile transforms to:

        Q_truncated(p) = Q_untruncated( p * (F_truncated(high) - F_truncated(low)) + F_truncated(low))

    And the CDF transforms to:

        F_truncated(x) = (F_untruncated(x) - F_untruncated(low)) / (F_untruncated(high) - F_untruncated(low))
    """

    def __init__(self, prior: AbstractPrior, low: Union[jax.Array, float], high: Union[jax.Array, float],
                 name: Optional[str] = None, base_dtype=jnp.float32):
        super(TruncationWrapper, self).__init__(name=name, base_dtype=base_dtype)
        self.prior = prior
        self.low = jnp.minimum(low, high)
        self.high = jnp.maximum(low, high)
        self.cdf_low = self.prior._inverse(self.low)
        self.cdf_diff = self.prior._inverse(self.high) - self.prior._inverse(self.low)

    def _inverse(self, X):
        return jnp.clip((self.prior._inverse(X) - self.cdf_low) / jnp.maximum(self.cdf_diff, 1e-6),
                        0., 1.)

    def _forward(self, U):
        return jnp.clip(self.prior._forward(jnp.clip(U * self.cdf_diff + self.cdf_low,
                                                     0., 1.)),
                        self.low, self.high)

    def _log_prob(self, X):
        outside_mask = jnp.bitwise_or(X < self.low, X > self.high)
        return jnp.where(outside_mask, -jnp.inf, self.prior._log_prob(X) - jnp.log(self.cdf_diff))

    def _base_shape(self):
        return self.prior._base_shape()

    def _shape(self):
        return self.prior._shape()

    def _dtype(self) -> jnp.dtype:
        return self.prior._dtype()


class ExplicitDensityPrior(AbstractPrior):
    def __init__(self, *, axes: Tuple[jax.Array, ...], density: jax.Array, regular_grid: bool = False,
                 name: Optional[str] = None, base_dtype=jnp.float32):
        super(ExplicitDensityPrior, self).__init__(name=name, base_dtype=base_dtype)
        self._num_dims = num_dims = len(np.shape(density))
        for i in range(num_dims):
            if len(np.shape(axes[i])) != 1:
                raise ValueError(f"Each axis must be 1D, got {np.shape(axes[i])}")
            if np.shape(density)[i] < 2:
                raise ValueError(f"Each dimension of density must have at least 2 elements, got {np.shape(density)}")
            if np.shape(density)[i] != np.size(axes[i]):
                raise ValueError(
                    f"Each dimension of density must have the same number of elements as the corresponding axis, "
                    f"got {np.shape(density)} and {np.shape(axes[i])}"
                )

        norm = density
        for i in range(num_dims)[::-1]:
            axis = axes[i]
            d = axis[1:] - axis[:-1]
            norm = 0.5 * jnp.sum((norm[..., :-1] + norm[..., 1:]) * d, axis=-1)

        self._density = density / norm
        self._axes = axes
        self._regular_grid = regular_grid

    def _dtype(self):
        return self._density.dtype

    def _base_shape(self):
        return (self._num_dims,)

    def _shape(self):
        return (self._num_dims,)

    def _forward(self, U):
        return self._quantile(U)

    def _inverse(self, X):
        return self._cdf(X)

    def _cdf(self, X):
        U = []
        for i in range(self._num_dims):
            # Marginalise the last dims
            density = self._density
            for j in range(i + 1, self._num_dims)[::-1]:
                axis = self._axes[j]
                d = axis[1:] - axis[:-1]
                density = 0.5 * jnp.sum((density[..., :-1] + density[..., 1:]) * d, axis=-1)

            # Now we have P(X_0, ..., X_i), interpolate to get norm * P(X_i | X_0, ...)

            for j in range(i):
                P = InterpolatedArray(
                    x=self._axes[j],
                    values=density,
                    axis=0,
                    regular_grid=self._regular_grid
                )
                density = P(X[j])  # P(X_1, ..., X_i)   [N1, ..., Ni]
            # Now we have P(X_i | X_0, ...), and we compute percentiles
            d = self._axes[i][1:] - self._axes[i][:-1]
            norm = 0.5 * jnp.sum((density[1:] + density[:-1]) * d)
            density /= norm
            percentiles = jnp.concatenate([jnp.asarray([0.]), jnp.cumsum(0.5 * (density[1:] + density[:-1]) * d)])
            u_i = jnp.interp(X[i], self._axes[i], percentiles)
            U.append(u_i)
        return jnp.stack(U, axis=-1)

    def _log_prob(self, X):
        density = self._density
        for i in range(self._num_dims):
            P = InterpolatedArray(
                x=self._axes[i],
                values=density,
                axis=0,
                regular_grid=self._regular_grid
            )
            density = P(X[i])
        return jnp.log(density)

    def _quantile(self, U):
        # Sequentially sample from: P(X_0), P(X_1 | X_0), P(X_2 | X_0, X_1), ...
        # P(A| B) = P(A, B) / sum_A P(B, A)
        X = []
        for i in range(self._num_dims):
            # Marginalise the last dims
            density = self._density
            for j in range(i + 1, self._num_dims)[::-1]:
                axis = self._axes[j]
                d = axis[1:] - axis[:-1]
                density = 0.5 * jnp.sum((density[..., :-1] + density[..., 1:]) * d, axis=-1)

            # Now we have P(X_0, ..., X_i), interpolate to get norm * P(X_i | X_0, ...)

            for j in range(i):
                P = InterpolatedArray(
                    x=self._axes[j],
                    values=density,
                    axis=0,
                    regular_grid=self._regular_grid
                )
                density = P(X[j])  # P(X_1, ..., X_i)   [N1, ..., Ni]
            # Now we have P(X_i | X_0, ...), and we compute percentiles
            d = self._axes[i][1:] - self._axes[i][:-1]
            norm = 0.5 * jnp.sum((density[1:] + density[:-1]) * d)
            density /= norm
            percentiles = jnp.concatenate([jnp.asarray([0.]), jnp.cumsum(0.5 * (density[1:] + density[:-1]) * d)])
            x_i = jnp.interp(U[i], percentiles, self._axes[i])
            X.append(x_i)
        return jnp.stack(X, axis=-1)


def apply_interp(x: jax.Array, i0: jax.Array, alpha0: jax.Array, i1: jax.Array, alpha1: jax, axis: int = 0):
    """
    Apply interpolation alpha given axis.

    Args:
        x: nd-array
        i0: [N] or scalar
        alpha0: [N] or scalar
        i1: [N] or scalar
        alpha1: [N] or scalar
        axis: axis to take along

    Returns:
        [N] or scalar interpolated along axis
    """

    def take(i):
        num_dims = len(np.shape(x))
        # [0] [1] [2 3 4], num_dims=5, axis=1
        slices = [slice(None)] * axis + [i] + [slice(None)] * (num_dims - axis - 1)
        # return jnp.take(x, i, axis=axis)
        return x[tuple(slices)]

    return left_broadcast_multiply(take(i0), alpha0, axis=axis) + left_broadcast_multiply(
        take(i1), alpha1, axis=axis)


def left_broadcast_multiply(x, y, axis: int = 0):
    """
    Left broadcast multiply of two arrays.
    Equivalent to right-padding before multiply

    Args:
        x: [..., a,b,c,...]
        y: [a, b]

    Returns:
        [..., a, b, c, ...]
    """
    needed_length = len(np.shape(x)[axis:])
    len_y = len(np.shape(y))
    extra = needed_length - len_y
    if extra < 0:
        raise ValueError(f"Shape mismatch {np.shape(x)} x {np.shape(y)}.")
    y = lax.reshape(y, np.shape(y) + (1,) * extra)
    return x * y


def get_interp_indices_and_weights(x, xp, regular_grid: bool = False):
    """
    One-dimensional linear interpolation. Outside bounds is also linear from nearest two points.

    Args:
        x: the x-coordinates at which to evaluate the interpolated values
        xp: the x-coordinates of the data points, must be increasing

    Returns:
        the interpolated values, same shape as `x`
    """

    if len(np.shape(xp)) == 0:
        xp = jnp.reshape(xp, (-1,))
    if np.shape(xp)[0] == 0:
        raise ValueError("xp must be non-empty")
    if np.shape(xp)[0] == 1:
        return (jnp.zeros_like(x, dtype=jnp.int32), jnp.ones_like(x)), (
            jnp.zeros_like(x, dtype=jnp.int32), jnp.zeros_like(x))

    # Find xp[i1-1] < x <= xp[i1]
    if regular_grid:
        # Use faster index determination
        delta_x = xp[1] - xp[0]
        i1 = jnp.clip((jnp.ceil((x - xp[0]) / delta_x)).astype(jnp.int64), 1, len(xp) - 1)
        i0 = i1 - 1
    else:
        i1 = jnp.clip(jnp.searchsorted(xp, x, side='right'), 1, len(xp) - 1)
        i0 = i1 - 1

    dx = xp[i1] - xp[i0]
    delta = x - xp[i0]

    epsilon = np.spacing(np.finfo(xp.dtype).eps)
    dx0 = jnp.abs(dx) <= epsilon  # Prevent NaN gradients when `dx` is small.
    dx = jnp.where(dx0, 1, dx)
    alpha = delta / dx
    return (i0, (1. - alpha)), (i1, alpha)


@dataclasses.dataclass(eq=False)
class InterpolatedArray:
    x: jax.Array  # [N]
    values: jax.Array  # [..., N, ...] `axis` has N elements

    axis: int = 0
    regular_grid: bool = False

    def __post_init__(self):

        if len(np.shape(self.x)) != 1:
            raise ValueError(f"Times must be 1D, got {np.shape(self.x)}.")

        def _assert_shape(x):
            if np.shape(x)[self.axis] != np.size(self.x):
                raise ValueError(f"Input values must have time length on `axis` dimension, got {np.shape(x)}.")

        jax.tree.map(_assert_shape, self.values)

        self.x, self.values = jax.tree.map(jnp.asarray, (self.x, self.values))

    @property
    def shape(self):
        return jax.tree.map(lambda x: np.shape(x)[:self.axis] + np.shape(x)[self.axis + 1:], self.values)

    def __call__(self, t: FloatArray) -> jax.Array:
        """
        Interpolate at time based on input times.

        Args:
            t: time to evaluate at.

        Returns:
            value at given time
        """
        (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(t, self.x, regular_grid=self.regular_grid)
        return jax.tree.map(lambda x: apply_interp(x, i0, alpha0, i1, alpha1, axis=self.axis), self.values)





from typing import Union

import jax
import numpy as np

__all__ = [
    'PRNGKey',
    'IntArray',
    'FloatArray',
    'BoolArray'
]

PRNGKey = jax.Array

Array = Union[
    jax.Array,  # JAX array type
    np.ndarray,  # NumPy array type
]
ComplexArray = Union[
    jax.Array,  # JAX array type
    np.ndarray,  # NumPy array type
    complex,  # valid scalars
]

FloatArray = Union[
    jax.Array,  # JAX array type
    np.ndarray,  # NumPy array type
    float,  # valid scalars
]
IntArray = Union[
    jax.Array,  # JAX array type
    np.ndarray,  # NumPy array type
    int,  # valid scalars
]
BoolArray = Union[
    jax.Array,  # JAX array type
    np.ndarray,  # NumPy array type
    np.bool_, bool,  # valid scalars
]

Array.__doc__ = "Type annotation for JAX array-like objects, with no scalar types."

ComplexArray.__doc__ = "Type annotation for JAX array-like objects, with complex scalar types."

FloatArray.__doc__ = "Type annotation for JAX array-like objects, with float scalar types."

IntArray.__doc__ = "Type annotation for JAX array-like objects, with int scalar types."

BoolArray.__doc__ = "Type annotation for JAX array-like objects, with bool scalar types."



import dataclasses
from collections import defaultdict
from contextlib import contextmanager
from functools import wraps
from typing import Any, Callable, Tuple, NamedTuple, Optional, TypeVar, List, Dict, Generic, Union, Iterator

import jax
from jax._src.typing import SupportsDType

from jaxctx.priors.types import PRNGKey

__all__ = [
    'get_parameter',
    'set_parameter',
    'transform',
    'convert_external_params',
    'wrap_random',
    'next_rng_key',
    'scope',
    'CtxParams',
    'ScopedDict'
]


class ScopedDict:
    """
    Nested dictionary keyed by scopes. Non-terminal nodes are scopes (dicts),
    terminal nodes are values.
    """

    def __init__(self, _dict=None, _scopes=None):
        self._scopes: List[str] = [] if _scopes is None else _scopes
        self._dict = {} if _dict is None else _dict

    def with_scopes(self, scopes: List[str]) -> 'ScopedDict':
        return ScopedDict(_dict=self._dict, _scopes=list(scopes))

    @property
    def scope_prefix(self):
        return '.'.join(self._scopes)

    def to_dict(self):
        return self._dict

    def __repr__(self):
        return f"ScopedDict(scopes={repr(self._scopes)}, dict={repr(self._dict)})"

    def _get_scope_dict(self, create: bool = False) -> Optional[Dict[str, Any]]:
        node = self._dict
        for scope in self._scopes:
            if scope not in node:
                if not create:
                    return None
                node[scope] = {}
            child = node[scope]
            if not isinstance(child, dict):
                raise ValueError(f"Scope '{scope}' collides with existing leaf.")
            node = child
        return node

    @staticmethod
    def _split_dotted(dotted_key: str) -> List[str]:
        if dotted_key.startswith('.'):
            dotted_key = dotted_key[1:]
        if dotted_key == "":
            return []
        parts = dotted_key.split('.')
        if any(part == "" for part in parts):
            raise ValueError("Dotted key contains empty scope segment.")
        return parts

    def get_dotted(self, dotted_key: str):
        """
        Lookup a value from the root using dotted notation.
        Leading '.' is optional and ignored.
        """
        parts = self._split_dotted(dotted_key)
        node = self._dict
        for part in parts:
            if not isinstance(node, dict) or part not in node:
                raise KeyError(dotted_key)
            node = node[part]
        return node

    def iter_items(self):
        """
        Iterate over leaf values yielding (dotted_key, value).
        Keys are returned relative to the root, prefixed by current scopes.
        """
        scope_dict = self._get_scope_dict(create=False)
        if scope_dict is None:
            return iter(())

        prefix = list(self._scopes)

        def _walk(node: Dict[str, Any], path_prefix: List[str]):
            for key in sorted(node.keys()):
                value = node[key]
                if isinstance(value, dict):
                    yield from _walk(value, path_prefix + [key])
                else:
                    dotted_key = '.'.join(path_prefix + [key])
                    yield dotted_key, value

        return _walk(scope_dict, prefix)

    def __getitem__(self, item):
        scope_dict = self._get_scope_dict(create=False)
        if scope_dict is None:
            raise KeyError(item)
        return scope_dict[item]

    def __setitem__(self, key, value):
        scope_dict = self._get_scope_dict(create=True)
        if scope_dict is None:
            raise ValueError("Scope path could not be resolved.")
        if key in scope_dict:
            existing = scope_dict[key]
            if isinstance(existing, dict) and not isinstance(value, dict):
                raise ValueError(f"Cannot overwrite scope '{key}' with a leaf value.")
            if not isinstance(existing, dict) and isinstance(value, dict):
                raise ValueError(f"Cannot overwrite leaf '{key}' with a scope.")
        scope_dict[key] = value

    def __contains__(self, item):
        scope_dict = self._get_scope_dict(create=False)
        if scope_dict is None:
            return False
        return item in scope_dict

    def __iter__(self):
        scope_dict = self._get_scope_dict(create=False)
        if scope_dict is None:
            return iter(())
        return iter(scope_dict)

    def __len__(self):
        scope_dict = self._get_scope_dict(create=False)
        if scope_dict is None:
            return 0
        return len(scope_dict)

    def keys(self):
        scope_dict = self._get_scope_dict(create=False) or {}
        return scope_dict.keys()

    def values(self):
        scope_dict = self._get_scope_dict(create=False) or {}
        return scope_dict.values()

    def items(self):
        scope_dict = self._get_scope_dict(create=False) or {}
        return scope_dict.items()


# Add as pytree type

def _flatten_mapping(mapping: Dict[str, Any]) -> Tuple[List[Any], Tuple[Tuple[str, Any], ...]]:
    children: List[Any] = []
    structure: List[Tuple[str, Any]] = []
    for key in sorted(mapping.keys()):
        value = mapping[key]
        if isinstance(value, dict):
            sub_children, sub_structure = _flatten_mapping(value)
            children.extend(sub_children)
            structure.append((key, sub_structure))
        else:
            children.append(value)
            structure.append((key, None))
    return children, tuple(structure)


def _unflatten_mapping(structure: Tuple[Tuple[str, Any], ...], children_iter: Iterator[Any]) -> Dict[str, Any]:
    mapping: Dict[str, Any] = {}
    for key, sub_structure in structure:
        if sub_structure is None:
            mapping[key] = next(children_iter)
        else:
            mapping[key] = _unflatten_mapping(sub_structure, children_iter)
    return mapping

def scoped_dict_flatten(scoped_dict: ScopedDict):
    children, structure = _flatten_mapping(scoped_dict._dict)
    return (
        children,
        (structure, tuple(scoped_dict._scopes))
    )


def scoped_dict_unflatten(aux_data, children):
    structure, scopes = aux_data
    children_iter = iter(children)
    _dict = _unflatten_mapping(structure, children_iter)
    return ScopedDict(_dict=_dict, _scopes=list(scopes))


jax.tree_util.register_pytree_node(
    ScopedDict,
    scoped_dict_flatten,
    scoped_dict_unflatten
)

CtxParams = ScopedDict


class Ctx:
    def __init__(self, rngs: Dict[str, PRNGKey], collections: Dict[str, ScopedDict], stack: List['Ctx'], init: bool):
        self._collections = defaultdict(ScopedDict)
        for key, val in collections.items():
            self._collections[key] = val
        self._rngs = rngs
        self._stack = stack
        self._scopes = []
        self._init = init

    def next_rng_key(self, rng_stream: str) -> PRNGKey:
        if rng_stream not in self._rngs:
            raise ValueError(f"RNG stream {rng_stream} not provided.")
        self._rngs[rng_stream], new_rng = jax.random.split(self._rngs[rng_stream])
        return new_rng

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stack.remove(self)
        return

    @property
    def is_init(self) -> bool:
        return self._init

    @property
    def collections(self) -> Dict[str, ScopedDict]:
        return dict(self._collections.items())

    def get_collection(self, collection: str) -> ScopedDict:
        return self._collections[collection].with_scopes(self._scopes)

    def push_scope(self, scope):
        self._scopes.append(scope)

    def pop_scope(self):
        self._scopes.pop()


class GlobalContext:
    def __init__(self, rng: Optional[jax.Array] = None):
        self.stack: List[Ctx] = []

    def new(self, rngs: Dict[str, PRNGKey], collections: Dict[str, ScopedDict], init: bool):
        new_ctx = Ctx(rngs=rngs, collections=collections, stack=self.stack, init=init)
        self.stack.append(new_ctx)
        return new_ctx

    @property
    def collections(self) -> Dict[str, ScopedDict]:
        if len(self.stack) == 0:
            raise ValueError("No context available. Must use `transform` to create a context.")
        return self.stack[-1].collections

    @property
    def is_init(self) -> bool:
        if len(self.stack) == 0:
            raise ValueError("No context available. Must use `transform` to create a context.")
        return self.stack[-1].is_init

    def get_collection(self, collection: str) -> ScopedDict:
        if len(self.stack) == 0:
            raise ValueError("No context available. Must use `transform` to create a context.")
        return self.stack[-1].get_collection(collection)

    def next_rng_key(self, rng_stream: str) -> PRNGKey:
        if len(self.stack) == 0:
            raise ValueError("No context available. Must use transform_with_state to create a context.")
        return self.stack[-1].next_rng_key(rng_stream)

    def push_scope(self, scope: str):
        if len(self.stack) == 0:
            raise ValueError("No context available. Must use transform_with_state to create a context.")
        self.stack[-1].push_scope(scope)

    def pop_scope(self):
        if len(self.stack) == 0:
            raise ValueError("No context available. Must use transform_with_state to create a context.")
        self.stack[-1].pop_scope()


global_context = GlobalContext()

PT = TypeVar('PT')
InitType = Union[jax.Array, Callable[[Tuple[int, ...], SupportsDType], PT], Callable[[], PT]]


@contextmanager
def scope(name: str):
    """
    Create a new scope, to prefix parameters and states, as {current_scope}.{name}.{param_name}.

    Args:
        name: the name of the scope

    Returns:
        The scope
    """
    # Context manager
    global_context.push_scope(name)
    try:
        yield
    finally:
        global_context.pop_scope()


def default_init(shape: Tuple[int, ...], dtype: SupportsDType):
    raise NotImplementedError("No init provided.")


def get_parameter(name: str, collection: str, shape: Optional[Tuple[int, ...]] = None,
                  dtype: Optional[SupportsDType] = None, *,
                  init: InitType = default_init) -> PT:
    """
    Get a parameter variable, initialised to a particular value. The parameter is stored in the given collection.

    Args:
        name: the name of the parameter
        collection: the collection to store parameter under.
        shape: the shape of the parameter must be provided if init is not a jax.Array
        dtype: the dtype of the parameter must be provided if init is not a jax.Array
        init: the initializer

    Returns:
        The parameter variable as a jax.Array
    """
    params = global_context.get_collection(collection)
    if name not in params:
        if callable(init):
            if (shape is None) and (dtype is None):
                new_param = init()
            else:
                new_param = init(shape, dtype)
        else:
            # Beware of closure issues.
            new_param = init
        params[name] = new_param

    return params[name]


def set_parameter(name: str, collection: str, value: PT) -> PT:
    """
    Set a variable in a collection to a particular value.
    If the name is not found in the collection then an error is raised.

    Args:
        name: the name of the state
        collection: the collection of the parameter
        value: the value to set

    Returns:
        The state variable as a jax.Array
    """
    params = global_context.get_collection(collection)
    if name not in params:
        raise ValueError(f"Parameter {name} not found. It must be initialised with `get_parameter` before it can be set.")
    # Ensure same pytree def
    tree_def = jax.tree.structure(params[name])
    value_tree_def = jax.tree.structure(value)
    if tree_def != value_tree_def:
        raise ValueError(f"Expected state with tree_def {tree_def} got {value_tree_def}.")
    if global_context.is_init:
        return params[name]
    params[name] = value
    return value


def set_state(name: str, collection: str, value: PT) -> PT:
    """
    Set a variable in a collection to a particular value.

    Args:
        name: the name of the state
        collection: the collection of the parameter
        value: the value to set

    Returns:
        The state variable as a jax.Array
    """
    params = global_context.get_collection(collection)
    if name in params:
        # Ensure same pytree def
        tree_def = jax.tree.structure(params[name])
        value_tree_def = jax.tree.structure(value)
        if tree_def != value_tree_def:
            raise ValueError(f"Expected state with tree_def {tree_def} got {value_tree_def}.")
    params[name] = value


def get_state(name: str, collection: str) -> PT:
    """
    Get a state in a collection.

    Args:
        name: the name of the state
        collection: the collection of the parameter

    Returns:
        The state variable as a jax.Array
    """
    params = global_context.get_collection(collection)
    if name not in params:
        raise ValueError(f"State {name} not found. It must be initialised before it can be set.")
    return params[name]


ExtParam = TypeVar('ExtParam')


def convert_external_params(external_params: ExtParam, collection: str, prefix: str = "") -> ExtParam:
    """
    Convert external parameters to context parameters. This can be used to convert haiku or flax parameters to
    jaxctx parameters.

    Args:
        external_params: map of name -> value
        collection: which collection to store the parameters in
        prefix: an optional prefix for the param name to make it unique.

    Returns:
        The context parameters
    """
    leaf_list, tree_def = jax.tree.flatten(external_params)

    def _unique_name(idx):
        return f"__{prefix}_{idx}"

    ctx_params = [get_parameter(_unique_name(idx), collection, init=leaf) for idx, leaf in enumerate(leaf_list)]
    external_params_ctx = jax.tree_unflatten(tree_def, ctx_params)
    return external_params_ctx


def wrap_random(f, rng_stream: str, **kwargs):
    """
    Wrap a function to use a random number generator from the context.
    Converts a function of (key, shape, dtype, **kwargs) into (shape, dtype).

    Args:
        f: the function to wrap, the args should be (key, shape, dtype, **kwargs)
        rng_stream: the rng stream to get from
        **kwargs options to pass to random generator.

    Returns:
        The wrapped function
    """

    @wraps(f)
    def wrapped(shape: Tuple[int, ...], dtype: SupportsDType):
        rng = next_rng_key(rng_stream)
        return f(rng, shape, dtype, **kwargs)

    return wrapped


FV = TypeVar('FV')


class ApplyReturn(NamedTuple, Generic[FV]):
    fn_val: FV
    collections: Dict[str, ScopedDict]


class InitReturn(NamedTuple, Generic[FV]):
    collections: Dict[str, ScopedDict]


@dataclasses.dataclass
class TransformedFn(Generic[FV]):
    _apply_fn: Callable
    _init_fn: Callable

    def apply(self, rngs, collections, *args, **kwargs) -> ApplyReturn[FV]:
        """
        Apply the function with given parameters and states.

        Args:
            rngs: a map of stream name to rng, e.g. {'params', jax.random.PRNGKey(0)}
            collections: a map of collection name to ScopedDict, e.g. {'params': ScopedDict()}
            *args: args to function
            **kwargs: kwargs to function

        Returns:
            The output of the function at the given input and the states
        """
        return self._apply_fn(rngs, collections, *args, **kwargs)

    def init(self, rngs, collections, *args, **kwargs) -> InitReturn[FV]:
        """
        Initialise the transform.

        Args:
            rngs: a map of stream name to rng, e.g. {'params', jax.random.PRNGKey(0)}
            collections: a map of collection name to ScopedDict, e.g. {'params': ScopedDict()}
            *args: args to function
            **kwargs: kwargs to function

        Returns:
            The output of the function at the given input and the states
        """
        return self._init_fn(rngs, collections, *args, **kwargs)


def transform(f: Callable[..., FV]) -> TransformedFn[FV]:
    """
    Transform a function that uses parameters and states into a pure function.

    Args:
        f: the function to transform

    Returns:
        A tuple of the init and apply functions
    """

    def apply_fn(rngs, collections, *args, **kwargs) -> ApplyReturn[FV]:
        """
        Apply the function with given parameters and states.

        Args:
            *args: args to function
            **kwargs: kwargs to function

        Returns:
            The output of the function at the given input and the states
        """
        if collections is None:
            collections = {}

        with global_context.new(rngs=rngs, collections=collections, init=False) as ctx:
            fn_val = f(*args, **kwargs)
            return ApplyReturn(fn_val=fn_val, collections=global_context.collections)

    def init_fn(rngs, collections, *args, **kwargs) -> InitReturn[FV]:
        """
        Initialise the transform.

        Args:
            *args: args to function
            **kwargs: kwargs to function

        Returns:
            The output of the function at the given input and the states
        """
        if collections is None:
            collections = {}

        with global_context.new(rngs=rngs, collections=collections, init=True) as ctx:
            # can be sped up with aeval
            _ = f(*args, **kwargs)
            del _  # Ensure no closure issues
            return InitReturn(collections=global_context.collections)

    return TransformedFn(_init_fn=init_fn, _apply_fn=apply_fn)


def next_rng_key(rng_stream: str):
    """
    Get the next random number generator

    Returns:
        The next random number generator
    """
    return global_context.next_rng_key(rng_stream)



import warnings

import jax
import numpy as np
import pytest
from jax import numpy as jnp, vmap

from jaxctx.priors.prior import Prior
from jaxctx.priors.special_priors import get_interp_indices_and_weights, apply_interp, left_broadcast_multiply, \
    InterpolatedArray, Bernoulli, Categorical, Poisson, Beta, ForcedIdentifiability, UnnormalisedDirichlet, \
    _poisson_quantile_bisection, _poisson_quantile, Empirical, tfpd, TruncationWrapper, ExplicitDensityPrior


def test_get_interp_indices_and_weights():
    xp = [0, 1, 2, 3]
    x = 1.5
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, jnp.asarray(xp, jnp.float32))
    print(i0, alpha0, i1, alpha1)
    assert i0 == 1
    assert alpha0 == 0.5
    assert i1 == 2
    assert alpha1 == 0.5

    x = 0
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, jnp.asarray(xp, jnp.float32))
    print(i0, alpha0, i1, alpha1)
    assert i0 == 0
    assert alpha0 == 1
    assert i1 == 1
    assert alpha1 == 0

    x = 3
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, jnp.asarray(xp, jnp.float32))
    print(i0, alpha0, i1, alpha1)
    assert i0 == 2
    assert alpha0 == 0
    assert i1 == 3
    assert alpha1 == 1

    x = -1
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, jnp.asarray(xp, jnp.float32))
    print(i0, alpha0, i1, alpha1)
    assert xp[i0] * alpha0 + xp[i1] * alpha1 == -1

    x = 4
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, jnp.asarray(xp, jnp.float32))
    print(i0, alpha0, i1, alpha1)
    assert xp[i0] * alpha0 + xp[i1] * alpha1 == 4

    x = 5
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, jnp.asarray(xp, jnp.float32))
    print(i0, alpha0, i1, alpha1)
    assert xp[i0] * alpha0 + xp[i1] * alpha1 == 5

    xp = [0., 0.]
    x = 0.
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, jnp.asarray(xp, jnp.float32))
    print(i0, alpha0, i1, alpha1)
    assert i0 == 0
    assert alpha0 == 1
    assert i1 == 1
    assert alpha1 == 0.

    xp = [0., 0.]
    x = -1
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, jnp.asarray(xp, jnp.float32))
    print(i0, alpha0, i1, alpha1)
    assert i0 == 0
    assert alpha0 == 2.
    assert i1 == 1
    assert alpha1 == -1.

    xp = [0.]
    x = 0.5
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, jnp.asarray(xp, jnp.float32))
    print(i0, alpha0, i1, alpha1)
    assert i0 == 0
    assert alpha0 == 1.
    assert i1 == 0
    assert alpha1 == 0.

    # Vector ops
    xp = [0, 1, 2, 3]
    x = [1.5, 1.5]
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(jnp.asarray(x, jnp.float32), jnp.asarray(xp, jnp.float32))
    print(i0, alpha0, i1, alpha1)
    np.testing.assert_array_equal(i0, jnp.asarray([1, 1]))
    np.testing.assert_array_equal(alpha0, jnp.asarray([0.5, 0.5]))
    np.testing.assert_array_equal(i1, jnp.asarray([2, 2]))
    np.testing.assert_array_equal(alpha1, jnp.asarray([0.5, 0.5]))

    xp = [0, 1, 2, 3]
    x = [1.5, 2.5]
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(jnp.asarray(x, jnp.float32), jnp.asarray(xp, jnp.float32))
    print(i0, alpha0, i1, alpha1)
    np.testing.assert_array_equal(i0, jnp.asarray([1, 2]))
    np.testing.assert_array_equal(alpha0, jnp.asarray([0.5, 0.5]))
    np.testing.assert_array_equal(i1, jnp.asarray([2, 3]))
    np.testing.assert_array_equal(alpha1, jnp.asarray([0.5, 0.5]))

    # xp = [0, 1, 2, 3]
    # x = [-0.5, 3.5]
    # (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, jnp.asarray(xp, jnp.float32))
    # print(i0, alpha0, i1, alpha1)
    # np.testing.assert_array_equal(i0, jnp.asarray([1, 2]))
    # np.testing.assert_array_equal(alpha0, jnp.asarray([0.5, 0.5]))
    # np.testing.assert_array_equal(i1, jnp.asarray([2, 3]))
    # np.testing.assert_array_equal(alpha1, jnp.asarray([0.5, 0.5]))


@pytest.mark.parametrize('regular_grid', [True, False])
def test_apply_interp(regular_grid):
    xp = jnp.linspace(0., 1., 10)
    x = jnp.linspace(-0.1, 1.1, 10)
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp, regular_grid=regular_grid)
    np.testing.assert_allclose(apply_interp(xp, i0, alpha0, i1, alpha1), x, atol=1e-6)

    x = jnp.linspace(-0.1, 1.1, 10)
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp, regular_grid=regular_grid)
    assert apply_interp(jnp.zeros((4, 5, 10, 6)), i0, alpha0, i1, alpha1, axis=2).shape == (4, 5, 10, 6)

    x = 0.
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp, regular_grid=regular_grid)
    assert apply_interp(jnp.zeros((4, 5, 10, 6)), i0, alpha0, i1, alpha1, axis=2).shape == (4, 5, 6)

    print(
        jax.jit(lambda x: apply_interp(x, i0, alpha0, i1, alpha1, axis=2)).lower(
            jnp.zeros((4, 5, 10, 6))).compile().cost_analysis()
    )
    # [{'bytes accessed': 1440.0, 'utilization1{}': 2.0, 'bytes accessed0{}': 960.0, 'bytes accessedout{}': 480.0, 'bytes accessed1{}': 960.0}]
    # [{'bytes accessed1{}': 960.0,  'utilization1{}': 2.0, 'bytes accessedout{}': 480.0, 'bytes accessed0{}': 960.0, 'bytes accessed': 1440.0}]


def test_regular_grid():
    # Inside bounds
    xp = jnp.linspace(0., 1., 10)
    fp = jax.random.normal(jax.random.PRNGKey(0), (10, 15))
    x = jnp.linspace(0., 1., 100)
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp, regular_grid=False)
    f_no = apply_interp(fp, i0, alpha0, i1, alpha1)

    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp, regular_grid=True)
    f_yes = apply_interp(fp, i0, alpha0, i1, alpha1)
    np.testing.assert_allclose(
        f_yes, f_no,
        atol=1e-6
    )

    # Outside bounds
    x = jnp.linspace(-0.1, 1.1, 100)
    xp = jnp.linspace(0., 1., 10)
    fp = jax.random.normal(jax.random.PRNGKey(0), (10, 15))
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp, regular_grid=False)
    f_no = apply_interp(fp, i0, alpha0, i1, alpha1)

    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp, regular_grid=True)
    f_yes = apply_interp(fp, i0, alpha0, i1, alpha1)
    np.testing.assert_allclose(
        f_yes, f_no,
        atol=1e-6
    )


def test_left_broadcast_multiply():
    assert np.all(left_broadcast_multiply(np.ones((2, 3)), np.ones((2,))) == np.ones((2, 3)))
    assert np.all(left_broadcast_multiply(np.ones((2, 3)), np.ones((2, 3))) == np.ones((2, 3)))
    assert np.all(
        left_broadcast_multiply(np.ones((1, 2, 3, 4, 5)), np.ones((3, 4)), axis=2) == np.ones((1, 2, 3, 4, 5)))
    assert np.all(
        left_broadcast_multiply(np.ones((1, 2, 3, 4, 5)), np.ones((3, 4)), axis=-3) == np.ones((1, 2, 3, 4, 5)))


@pytest.mark.parametrize('regular_grid', [True, False])
def test_interpolated_array(regular_grid: bool):
    # scalar time
    times = jnp.linspace(0, 10, 100)
    values = jnp.sin(times)
    interp = InterpolatedArray(times, values, regular_grid=regular_grid)
    assert interp(5.).shape == ()
    np.testing.assert_allclose(interp(5.), jnp.sin(5), atol=2e-3)

    # vector time
    assert interp(jnp.array([5., 6.])).shape == (2,)
    np.testing.assert_allclose(interp(jnp.array([5., 6.])), jnp.sin(jnp.array([5., 6.])), atol=2e-3)

    # Now with axis = 1
    times = jnp.linspace(0, 10, 100)
    values = jnp.stack([jnp.sin(times), jnp.cos(times)], axis=0)  # [2, 100]
    interp = InterpolatedArray(times, values, axis=1, regular_grid=regular_grid)
    assert interp(5.).shape == (2,)
    np.testing.assert_allclose(interp(5.), jnp.array([jnp.sin(5), jnp.cos(5)]), atol=2e-3)

    # Vector
    assert interp(jnp.array([5., 6., 7.])).shape == (2, 3)
    np.testing.assert_allclose(interp(jnp.array([5., 6., 7.])),
                               jnp.stack([jnp.sin(jnp.array([5., 6., 7.])), jnp.cos(jnp.array([5., 6., 7.]))],
                                         axis=0),
                               atol=2e-3)


@pytest.fixture(scope='module')
def mock_special_priors():
    # prior, (min, max), shape
    return [
        (Bernoulli(probs=jnp.ones(5), name='x'), (0, 1), (5,)),
        (Categorical(parametrisation='gumbel_max', probs=jnp.ones(5), name='x'), (0, 4), ()),
        (Categorical(parametrisation='cdf', probs=jnp.ones(5), name='x'), (0, 4), ()),
        (Poisson(rate=jnp.ones(5), name='x'), (0, 100), (5,)),
        (Beta(concentration0=jnp.ones(5), concentration1=jnp.ones(5), name='x'), (0, 1), (5,)),
        (ForcedIdentifiability(n=10, low=jnp.zeros(5), high=jnp.ones(5), name='x'), (0, 1), (10, 5)),
        (ForcedIdentifiability(n=10, low=jnp.zeros(5), high=jnp.ones(5), fix_left=True, name='x'), (0, 1), (10, 5)),
        (ForcedIdentifiability(n=10, low=jnp.zeros(5), high=jnp.ones(5), fix_right=True, name='x'), (0, 1), (10, 5)),
        (ForcedIdentifiability(n=10, low=jnp.zeros(5), high=jnp.ones(5), fix_left=True, fix_right=True, name='x'),
         (0, 1), (10, 5)),
        (UnnormalisedDirichlet(concentration=jnp.ones(5), name='x'), (0, jnp.inf), (5,)),
        (UnnormalisedDirichlet(concentration=jnp.ones((3, 5)), name='x'), (0, jnp.inf), (3, 5)),
    ]


def test_special_priors(mock_special_priors):
    for prior, (vmin, vmax), shape in mock_special_priors:
        print(f"Testing {prior.__class__}")

        x = prior.forward(jnp.ones(prior.base_shape, prior.base_dtype))
        assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
        assert jnp.all(x >= vmin)
        assert jnp.all(x <= vmax)
        assert x.shape == shape
        assert x.shape == prior.shape
        x = prior.forward(jnp.zeros(prior.base_shape, prior.base_dtype))
        assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
        assert jnp.all(x >= vmin)
        assert jnp.all(x <= vmax)
        assert x.shape == shape
        assert x.shape == prior.shape

        u_input = vmap(lambda key: jax.random.uniform(key, shape=prior.base_shape))(jax.random.split(jax.random.PRNGKey(42), 10))
        x = vmap(lambda u: prior.forward(u))(u_input)
        assert jnp.all(x >= vmin)
        assert jnp.all(x <= vmax)

        try:
            u = vmap(lambda x: prior.inverse(x))(x)
            assert u.shape[1:] == prior.base_shape

            if prior.dtype in [jnp.bool_, jnp.int32, jnp.int64]:
                continue
            assert jnp.allclose(u, u_input)
        except NotImplementedError:
            warnings.warn(f"Skipping inverse test for {prior.__class__}")
            pass


@pytest.mark.parametrize("rate, error", (
        [2.0, 1.],
        [10., 1.],
        [100., 1.],
        [1000., 1.],
        [10000., 1.]
)
                         )
def test_poisson_quantile_bisection(rate, error):
    U = jnp.linspace(0., 1. - np.spacing(1.), 1000)
    x, x_results = _poisson_quantile_bisection(U, rate, unroll=False)
    diff_last_two = jnp.abs(x_results[..., -1] - x_results[..., -2])

    # Make sure less than 1 apart
    assert jnp.all(diff_last_two <= error)


@pytest.mark.parametrize("rate", [2.0, 10., 100., 1000., 10000.])
def test_poisson_quantile(rate):
    U = jnp.linspace(0., 1. - np.spacing(1.), 10000)
    x = _poisson_quantile(U, rate)
    assert jnp.all(jnp.isfinite(x))


def test_forced_identifiability():
    x = jnp.asarray([0., 0.1, 0.2, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., name='x')
    u = prior.inverse(x)
    print(u)
    assert jnp.allclose(prior.forward(u), x)

    x = jnp.asarray([0., 0.11, 0.1, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., name='x')
    u = prior.inverse(x)
    print(u)
    print(prior.forward(u))
    assert jnp.allclose(prior.forward(u), x)

    x = jnp.asarray([0., 0.1, 0.2, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., fix_right=True, name='x')
    u = prior.inverse(x)
    print(u)
    assert jnp.allclose(prior.forward(u), x)

    x = jnp.asarray([0., 0.11, 0.1, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., fix_right=True, name='x')
    u = prior.inverse(x)
    print(u)
    print(prior.forward(u))
    assert jnp.allclose(prior.forward(u), x)

    x = jnp.asarray([0., 0.1, 0.2, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., fix_left=True, name='x')
    u = prior.inverse(x)
    print(u)
    assert jnp.allclose(prior.forward(u), x)

    x = jnp.asarray([0., 0.11, 0.11, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., fix_left=True, name='x')
    u = prior.inverse(x)
    print(u)
    print(prior.forward(u))
    assert jnp.allclose(prior.forward(u), x)

    x = jnp.asarray([0., 0.1, 0.2, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., fix_left=True, fix_right=True, name='x')
    u = prior.inverse(x)
    print(u)
    assert jnp.allclose(prior.forward(u), x)

    x = jnp.asarray([0., 0.11, 0.1, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., fix_left=True, fix_right=True, name='x')
    u = prior.inverse(x)
    print(u)
    print(prior.forward(u))
    assert jnp.allclose(prior.forward(u), x)


def test_empirical():
    samples = jax.random.normal(jax.random.PRNGKey(42), shape=(2000,), dtype=jnp.float32)
    prior = Empirical(samples=samples, resolution=100, name='x')
    assert prior._percentiles.shape == (101, 1)

    x = prior.forward(jnp.ones(prior.base_shape, prior.base_dtype))
    assert x.shape == ()
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    x = prior.forward(jnp.zeros(prior.base_shape, prior.base_dtype))
    assert x.shape == ()
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))

    x = prior.forward(0.5 * jnp.ones(prior.base_shape, prior.base_dtype))
    np.testing.assert_allclose(x, 0., atol=0.06)

    u_input = vmap(lambda key: jax.random.uniform(key, shape=prior.base_shape, dtype=prior.base_dtype))(
        jax.random.split(jax.random.PRNGKey(42), 1000))
    x = vmap(lambda u: prior.forward(u))(u_input)
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))

    u = vmap(lambda x: prior.inverse(x))(x)
    # print(u)
    np.testing.assert_allclose(u, u_input, atol=5e-7)
    assert u.shape[1:] == prior.base_shape


def test_truncation_wrapper():
    prior = Prior(tfpd.Normal(loc=jnp.zeros(5), scale=jnp.ones(5)))
    trancated_prior = TruncationWrapper(prior=prior, low=0., high=1.)

    x = trancated_prior.forward(jnp.ones(trancated_prior.base_shape, prior.base_dtype))
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= 0.)
    assert jnp.all(x <= 1.)
    assert x.shape == (5,)

    x = trancated_prior.forward(jnp.zeros(trancated_prior.base_shape, prior.base_dtype))
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= 0.)
    assert jnp.all(x <= 1.)
    assert x.shape == (5,)

    u_input = vmap(lambda key: jax.random.uniform(key, shape=trancated_prior.base_shape))(
        jax.random.split(jax.random.PRNGKey(42), 1000))
    x = vmap(lambda u: trancated_prior.forward(u))(u_input)
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= 0.)
    assert jnp.all(x <= 1.)

    u = vmap(lambda x: trancated_prior.inverse(x))(x)
    np.testing.assert_allclose(u, u_input, atol=5e-7)

    prior = Prior(tfpd.Normal(loc=jnp.zeros(5), scale=jnp.ones(5)))
    trancated_prior = TruncationWrapper(prior=prior, low=-jnp.inf, high=1.)

    x = trancated_prior.forward(jnp.ones(trancated_prior.base_shape, prior.base_dtype))
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= -jnp.inf)
    assert jnp.all(x <= 1.)
    assert x.shape == (5,)

    x = trancated_prior.forward(jnp.zeros(trancated_prior.base_shape, prior.base_dtype))
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= -jnp.inf)
    assert jnp.all(x <= 1.)
    assert x.shape == (5,)

    u_input = vmap(lambda key: jax.random.uniform(key, shape=trancated_prior.base_shape))(
        jax.random.split(jax.random.PRNGKey(42), 1000))
    x = vmap(lambda u: trancated_prior.forward(u))(u_input)
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= -jnp.inf)
    assert jnp.all(x <= 1.)

    u = vmap(lambda x: trancated_prior.inverse(x))(x)
    np.testing.assert_allclose(u, u_input, atol=5e-7)

    prior = Prior(tfpd.Normal(loc=jnp.zeros(5), scale=0.01 * jnp.ones(5)))
    trancated_prior = TruncationWrapper(prior=prior, low=0., high=1.)

    x = trancated_prior.forward(jnp.ones(trancated_prior.base_shape, prior.base_dtype))
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= 0.)
    assert jnp.all(x <= 1.)
    assert x.shape == (5,)

    x = trancated_prior.forward(jnp.zeros(trancated_prior.base_shape, prior.base_dtype))
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= 0.)
    assert jnp.all(x <= 1.)
    assert x.shape == (5,)

    u_input = vmap(lambda key: jax.random.uniform(key, shape=trancated_prior.base_shape))(
        jax.random.split(jax.random.PRNGKey(42), 1000))
    x = vmap(lambda u: trancated_prior.forward(u))(u_input)
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= 0.)
    assert jnp.all(x <= 1.)

    u = vmap(lambda x: trancated_prior.inverse(x))(x)
    np.testing.assert_allclose(u, u_input, atol=5e-7)


def test_explicit_density_prior():
    resolution = 10
    density = jnp.ones((resolution + 1, resolution))
    axes = (jnp.linspace(0, 1, resolution + 1), jnp.linspace(0, 1, resolution))
    prior = ExplicitDensityPrior(axes=axes, density=density, regular_grid=True)

    x = prior.forward(jnp.ones(prior.base_shape, prior.base_dtype))
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x == 1.)
    assert x.shape == (2,)

    x = prior.forward(jnp.zeros(prior.base_shape, prior.base_dtype))
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x == 0.)
    assert x.shape == (2,)

    u_input = vmap(lambda key: jax.random.uniform(key, shape=prior.base_shape))(jax.random.split(jax.random.PRNGKey(42), 1000))
    x = vmap(lambda u: prior.forward(u))(u_input)
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= 0.)
    assert jnp.all(x <= 1.)

    u = vmap(lambda x: prior.inverse(x))(x)
    np.testing.assert_allclose(u, u_input, atol=5e-7)

    assert jnp.all(jnp.isfinite(jax.vmap(prior.log_prob)(x)))



import jax
import numpy as np
from jax import numpy as jnp

from jaxctx.context import get_parameter, wrap_random, set_parameter, transform, ScopedDict, scope


def test_transform():
    with jax.checking_leaks():
        def f(x) -> jax.Array:
            y = get_parameter(
                'y', 'params', (), jnp.float32,
                init=wrap_random(jax.random.normal, 'params')
            )
            s = get_parameter('s', 'state', y.shape, y.dtype, init=jnp.zeros)
            s = set_parameter('s', 'state', s + x + y)
            return s

        transformed = transform(f)

        init = jax.jit(transformed.init)({'params': jax.random.PRNGKey(0)}, {}, 1)
        print(init)

        apply = jax.jit(transformed.apply)

        response = apply({'params': jax.random.PRNGKey(0)}, init.collections, 1)
        print(response)
        assert response.fn_val == 1 + response.collections['params']['y']
        assert response.fn_val == response.collections['state']['s']
        for key, val in response.collections.items():
            assert isinstance(val, ScopedDict)

        next_response = apply({'params': jax.random.PRNGKey(0)}, response.collections, 1)

        print(next_response)
        assert next_response.fn_val == response.collections['state']['s'] + 1 + response.collections['params']['y']
        assert next_response.fn_val == next_response.collections['state']['s']


def test_scoped_dict_nested_and_dotted():
    with jax.checking_leaks():
        def f(x) -> jax.Array:
            with scope('layer'):
                y = get_parameter(
                    'y', 'params', (), jnp.float32,
                    init=wrap_random(jax.random.normal, 'params')
                )
                s = get_parameter('s', 'state', y.shape, y.dtype, init=jnp.zeros)
                s = set_parameter('s', 'state', s + x + y)
            return s

        transformed = transform(f)

        init = jax.jit(transformed.init)({'params': jax.random.PRNGKey(0)}, {}, 1)
        apply = jax.jit(transformed.apply)

        response = apply({'params': jax.random.PRNGKey(0)}, init.collections, 1)
        assert response.fn_val == 1 + response.collections['params']['layer']['y']
        assert response.fn_val == response.collections['state']['layer']['s']

        params = response.collections['params']
        np.testing.assert_array_equal(
            params.get_dotted('.layer.y'),
            response.collections['params']['layer']['y']
        )

        items = dict(params.iter_items())
        assert 'layer.y' in items
        np.testing.assert_array_equal(
            items['layer.y'],
            response.collections['params']['layer']['y']
        )



import jax
import numpy as np
import pytest
from jax import numpy as jnp
from jax.scipy import special as jsp

from jaxctx import transform
from jaxctx.priors.prior import quick_unit, quick_unit_inverse, Prior, tfpd, distribution_chain


def test_quick_unit():
    x = jnp.linspace(-10, 10, 1000000)
    y = quick_unit(x)
    assert np.all(y <= 1)
    assert np.all(y >= 0)
    x_reconstructed = quick_unit_inverse(y)
    np.testing.assert_allclose(x, x_reconstructed, atol=2e-5)

    g = jax.grad(quick_unit)
    assert np.all(np.isfinite(jax.vmap(g)(x)))
    assert np.isfinite(g(0.))

    h = jax.grad(quick_unit_inverse)
    assert np.all(np.isfinite(jax.vmap(h)(y)))
    assert np.isfinite(h(0.5))

    # Test performance against sigmoid and logit
    import time
    for f in [quick_unit, jax.nn.sigmoid]:
        g = jax.jit(f).lower(x).compile()
        t0 = time.time()
        for _ in range(1000):
            g(x).block_until_ready()
        print(f"{f.__name__} {time.time() - t0}s")

    for f in [quick_unit_inverse, jsp.logit]:
        g = jax.jit(f).lower(y).compile()
        t0 = time.time()
        for _ in range(1000):
            g(y).block_until_ready()
        print(f"{f.__name__} {time.time() - t0}s")


def _tail_threshold(forward, inverse, dtype, max_x: float = 1e4, iters: int = 80) -> float:
    def _is_valid(x_value: float) -> bool:
        x_arr = jnp.asarray(x_value, dtype=dtype)
        y = forward(x_arr)
        x_rec = inverse(y)
        return bool((y > 0) & (y < 1) & jnp.isfinite(y) & jnp.isfinite(x_rec))

    lo = 0.0
    hi = 1.0
    while hi < max_x and _is_valid(hi):
        lo = hi
        hi *= 2.0

    if _is_valid(hi):
        return float(hi)

    for _ in range(iters):
        mid = 0.5 * (lo + hi)
        if _is_valid(mid):
            lo = mid
        else:
            hi = mid
    return float(lo)


def test_quick_unit_tail_saturation():
    jax.config.update("jax_enable_x64", True)
    assert jax.config.read("jax_enable_x64")

    thresholds = {}
    for dtype in (jnp.float32, jnp.float64):
        thresholds[(dtype, "sigmoid")] = _tail_threshold(
            jax.nn.sigmoid,
            jsp.logit,
            dtype
        )
        thresholds[(dtype, "ndtr")] = _tail_threshold(
            jsp.ndtr,
            jsp.ndtri,
            dtype
        )
        print(f"{dtype} sigmoid tail threshold: {thresholds[(dtype, 'sigmoid')]}")
        print(f"{dtype} ndtr tail threshold: {thresholds[(dtype, 'ndtr')]}")

        assert thresholds[(dtype, "sigmoid")] > 0
        assert thresholds[(dtype, "ndtr")] > 0
        assert thresholds[(dtype, "sigmoid")] >= thresholds[(dtype, "ndtr")]

    assert thresholds[(jnp.float64, "sigmoid")] >= thresholds[(jnp.float32, "sigmoid")]
    assert thresholds[(jnp.float64, "ndtr")] >= thresholds[(jnp.float32, "ndtr")]


def test_prior():
    def model():
        x = Prior(tfpd.Normal(loc=0., scale=1.), name='x').realise()
        y = Prior(tfpd.Uniform(low=0., high=1.), name='y').parameter()
        y_rand = Prior(tfpd.Uniform(low=0., high=1.), name='y_rand').parameter(random_init=True)
        y_init = Prior(tfpd.Uniform(low=0., high=1.), name='y_init').parameter(init=0.75)
        y_init_callable = Prior(tfpd.Uniform(low=0., high=1.), name='y_init_callable').parameter(
            init=lambda key, shape, dtype: 0.75 * jnp.ones(shape, dtype))
        z = Prior(tfpd.Beta(concentration0=0.5, concentration1=1.), name='z').realise()
        return x, y, y_rand, y_init, y_init_callable, z

    transformed_model = transform(model)
    params = transformed_model.init({'params': jax.random.PRNGKey(0), 'U': jax.random.PRNGKey(1)}, {}).collections
    print(params)
    y_init = params['X']['y_init']
    assert np.isclose(y_init, 0.75, atol=1e-5)
    y_init_callable = params['X']['y_init_callable']
    assert np.isclose(y_init_callable, 0.75, atol=1e-5)

    print(transformed_model.apply({}, params))


def test_no_quantile_prior():
    def prior_model():
        z = Prior(tfpd.VonMises(loc=0., concentration=1.)).realise()
        return z

    with pytest.raises(ValueError):
        transform(prior_model).init({'params': jax.random.PRNGKey(0)}, {})


def test_distribution_chain():
    d = tfpd.MultivariateNormalTriL(loc=jnp.zeros(5), scale_tril=jnp.eye(5))
    chain = distribution_chain(d)
    assert len(chain) == 2
    assert isinstance(chain[0], tfpd.Sample)
    assert isinstance(chain[1], tfpd.MultivariateNormalTriL)

    chain = distribution_chain(tfpd.Normal(loc=jnp.zeros(5), scale=jnp.ones(5)))
    assert len(chain) == 1
    assert isinstance(chain[0], tfpd.Normal)


def test_priors():
    d = Prior(tfpd.Uniform(low=jnp.zeros(5), high=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, jnp.float32)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, jnp.float32)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Normal(loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, jnp.float32)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Laplace(loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, jnp.float32)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, jnp.float32)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Cauchy(loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, jnp.float32)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, jnp.float32)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.StudentT(df=1.5, loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, jnp.float32)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, jnp.float32)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Beta(concentration0=jnp.ones(5), concentration1=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, jnp.float32)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, jnp.float32)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.HalfNormal(scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, jnp.float32)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, jnp.float32)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.HalfCauchy(loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, jnp.float32)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, jnp.float32)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Gamma(concentration=jnp.ones(5), rate=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, jnp.float32)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, jnp.float32)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Gumbel(loc=jnp.ones(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, jnp.float32)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, jnp.float32)).shape == d.shape
    assert d.base_shape == ()
    assert d.shape == (5,)

    d = Prior(tfpd.MultivariateNormalTriL(loc=jnp.zeros(5), scale_tril=jnp.eye(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, jnp.float32)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, jnp.float32)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)


def test_various_collections():
    # We want to be able to create a model with
    def model():
        x = Prior(tfpd.Normal(loc=0., scale=1.), name='x').parameter()
        y = Prior(tfpd.Uniform(low=0., high=1.), name='y').realise()
        return x, y

    transformed_model = transform(model)
    params = transformed_model.init({'params': jax.random.PRNGKey(0), 'U': jax.random.PRNGKey(1)}, None).collections
    print(params)

    response = transformed_model.apply({'params': jax.random.PRNGKey(2), 'U': jax.random.PRNGKey(3)}, params)
    print(response)



import warnings
from abc import ABC, abstractmethod
from functools import partial
from typing import Tuple, Optional, Union, List, Callable

import jax
import numpy as np
import tensorflow_probability.substrates.jax as tfp
from jax import numpy as jnp

from jaxctx import wrap_random, get_parameter
from jaxctx.priors.types import FloatArray, IntArray, BoolArray, ComplexArray

tfpd = tfp.distributions

QUICK_UNIT_SCALE = 2.2

CoDomain = ComplexArray | FloatArray | IntArray | BoolArray


class AbstractPrior(ABC):
    """
    Represents a generative prior.
    """

    def __init__(self, name: str | None, base_dtype):
        self._name = name
        self._base_dtype = base_dtype

    def __repr__(self):
        return f"{self.name if self.name is not None else '*'}\t{self.base_shape} -> {self.shape} {self.dtype}"

    @property
    def name(self):
        """
        The name of the prior.
        """
        return self._name

    @abstractmethod
    def _dtype(self):
        """
        The dtype of the prior.
        """
        ...

    @abstractmethod
    def _base_shape(self) -> Tuple[int, ...]:
        """
        The base shape of the prior, in U-space.
        """
        ...

    @abstractmethod
    def _shape(self) -> Tuple[int, ...]:
        """
        The shape of the prior, in X-space.
        """
        ...

    @abstractmethod
    def _forward(self, U: FloatArray) -> CoDomain:
        """
        The forward transformation from U-space to X-space.

        Args:
            U: U-space representation

        Returns:
            X-space representation
        """
        ...

    @abstractmethod
    def _inverse(self, X: CoDomain) -> FloatArray:
        """
        The inverse transformation from X-space to U-space.

        Args:
            X: X-space representation

        Returns:
            U-space representation
        """
        ...

    @abstractmethod
    def _log_prob(self, X: CoDomain) -> FloatArray:
        """
        The log probability of the prior.

        Args:
            X: X-space representation

        Returns:
            log probability of the prior
        """

        ...

    @property
    def dtype(self):
        """
        The dtype of the prior random variable in X-space.
        """
        return self._dtype()

    @property
    def base_dtype(self):
        """
        The dtype of the prior random variable in X-space.
        """
        return self._base_dtype

    @property
    def base_shape(self) -> Tuple[int, ...]:
        """
        The base shape of the prior random variable in U-space.
        """
        return self._base_shape()

    @property
    def base_ndims(self):
        """
        The number of dimensions of the prior random variable in U-space.
        """
        return int(np.prod(self.base_shape))

    @property
    def shape(self) -> Tuple[int, ...]:
        """
        The shape of the prior random variable in X-space.
        """
        return self._shape()

    def forward(self, U: FloatArray) -> CoDomain:
        """
        The forward transformation from U-space to X-space.

        Args:
            U: U-space representation

        Returns:
            X-space representation
        """
        return self._forward(U)

    def inverse(self, X: CoDomain) -> FloatArray:
        """
        The inverse transformation from X-space to U-space.

        Args:
            X: X-space representation

        Returns:
            U-space representation
        """
        return self._inverse(X)

    def log_prob(self, X: CoDomain) -> FloatArray:
        """
        The log probability of the prior.

        Args:
            X: X-space representation

        Returns:
            log probability of the prior
        """
        log_prob = self._log_prob(X)
        if np.size(log_prob) > 1:
            log_prob = jnp.sum(log_prob)
        if log_prob.shape != ():
            log_prob = jax.lax.reshape(log_prob, ())
        return log_prob

    def parameter(self, *, init: FloatArray | Callable|None = None, random_init=False, param_collection: str = 'params',
                  U_collection: str = 'U', X_collection='X', log_prob_collection='log_prob',
                  rng_stream: str = 'params'):
        """
        Convert a prior into a constrained parameter, that takes a single value in the model, but still has an associated
        log_prob. The parameter is registered into the corresponding collection and stream.

        Intent is to use these for optimisation.

        Args:
            init: optional initial value or callable, giving the initial value in X-space.
            random_init: Whether to initialise the parameter randomly or at the median of the distribution.
            param_collection: The collection to register the parameter in.
            rng_stream: The name of the random number generator stream to use for sampling.

        Returns:
            a parameter constrained to the prior distribution.
        """
        return prior_to_parameter(prior=self, init=init, random_init=random_init,
                                  param_collection=param_collection,
                                  U_collection=U_collection, X_collection=X_collection,
                                  log_prob_collection=log_prob_collection,
                                  rng_stream=rng_stream)

    def realise(self, *, U_collection: str = 'U', X_collection: str = 'X', log_prob_collection='log_prob',
                rng_stream: str = 'U'):
        """
        Realise the prior distribution into a parameter.

        Args:
            U_collection: The collection to register the parameter in.
            X_collection: The collection to register the parameter in.
            rng_stream: The name of the random number generator stream to use for sampling U.

        Returns:
            A parameter representing the prior.
        """
        return realise_prior(prior=self, U_collection=U_collection, X_collection=X_collection,
                             log_prob_collection=log_prob_collection, rng_stream=rng_stream)


class Prior(AbstractPrior):
    """
    Represents a generative prior.
    """

    def __init__(self, dist: tfpd.Distribution, name: Optional[str] = None, base_dtype=jnp.float32):
        AbstractPrior.__init__(self, name=name, base_dtype=base_dtype)
        self._dist_chain = TFPDistributionChain(dist)
        self._dist = dist

    @property
    def dist(self) -> tfpd.Distribution:
        """
        The distribution of the prior.
        """
        return self._dist

    def _base_shape(self) -> Tuple[int, ...]:
        return self._dist_chain.base_shape()

    def _shape(self) -> Tuple[int, ...]:
        return self._dist_chain.shape()

    def _dtype(self):
        return self._dist_chain.dtype()

    def _forward(self, U: FloatArray) -> CoDomain:
        return self._dist_chain.forward(U)

    def _inverse(self, X: CoDomain) -> FloatArray:
        return self._dist_chain.inverse(X)

    def _log_prob(self, X: CoDomain) -> FloatArray:
        return self._dist_chain.log_prob(X=X)


def distribution_chain(dist: tfpd.Distribution) -> List[tfpd.TransformedDistribution | tfpd.Sample | tfpd.Distribution]:
    """
    Returns a list of distributions that make up the chain of distributions.

    Args:
        dist: A TFP distribution, transformed distribution or sample.

    Returns:
        A list of distributions.
    """
    chain = []
    while True:
        chain.append(dist)
        if isinstance(dist, tfpd.TransformedDistribution):
            dist = dist.distribution
            continue
        break
    # Must reverse the chain because the first distribution is the last in the chain.
    return chain[::-1]


class TFPDistributionChain:
    """
    Represents a wrapped TFP distribution.
    """

    def __init__(self, dist: tfpd.Distribution):
        self.dist_chain = distribution_chain(dist)
        check_dist = self.dist_chain[0]
        if isinstance(self.dist_chain[0], tfpd.Sample):
            check_dist = self.dist_chain[0].distribution
        if '_quantile' not in check_dist.__class__.__dict__:
            # TODO(Joshuaalbert): we could numerically approximate it. This requires knowing the support of dist.
            # Repartitioning the prior also requires knowing the support and choosing a replacement, which is not
            # always easy from stats. E.g. StudentT variance doesn't exist but a numerial quantile can be made.
            raise ValueError(f"Distribution {dist} is missing a quantile.")

    def __repr__(self):
        return " -> ".join(map(repr, self.dist_chain))

    def dtype(self):
        return self.dist_chain[-1].dtype

    def base_shape(self) -> Tuple[int, ...]:
        return tuple(self.dist_chain[0].batch_shape_tensor()) + tuple(self.dist_chain[0].event_shape_tensor())

    def shape(self) -> Tuple[int, ...]:
        return tuple(self.dist_chain[-1].batch_shape_tensor()) + tuple(self.dist_chain[-1].event_shape_tensor())

    def forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        dist = self.dist_chain[0]
        # cast U to the correct dtype for the distribution
        U = U.astype(jnp.result_type(U.dtype, dist.dtype))
        if isinstance(dist, tfpd.Sample):
            dist = dist.distribution
        X = dist.quantile(U)
        for dist in self.dist_chain[1:]:
            X = dist.bijector.forward(X)
        return X

    def inverse(self, X) -> FloatArray:
        for dist in reversed(self.dist_chain[1:]):
            X = dist.bijector.inverse(X)
        dist = self.dist_chain[0]
        if isinstance(dist, tfpd.Sample):
            dist = dist.distribution
        X = dist.cdf(X)
        return X

    def log_prob(self, X):
        return self.dist_chain[-1].log_prob(X)


def quick_unit(x: jax.Array) -> jax.Array:
    """
    Quick approximation to the sigmoid.

    Args:
        x: jax.Array value in (-inf, inf) open interval

    Returns:
        value in (0, 1) in open interval
    """
    return jax.nn.sigmoid(x / QUICK_UNIT_SCALE)


def quick_unit_inverse(y: jax.Array) -> jax.Array:
    """
    Inverse of quick_unit.

    Args:
        y: jax.Array value in (0, 1) open interval

    Returns:
        value in (-inf, inf) in open interval
    """
    return QUICK_UNIT_SCALE * jax.scipy.special.logit(y)


def sample_quick_unit_dist(key, shape, dtype):
    """
    Sample from the quick unit distribution in the param space, a logit-normal distribution with mean 0 and scale
    QUICK_UNIT_SCALE.

    Args:
        key: PRNGKey to use.
        shape: Shape of the output.
        dtype: Dtype of the output.

    Returns:
        A jax.Array sampled from the quick unit distribution.
    """

    return jax.random.normal(key, shape, dtype)


def prior_to_parameter(prior: AbstractPrior, init: FloatArray | Callable | None = None, random_init: bool = False,
                       param_collection: str = 'params',
                       U_collection: str = 'U', X_collection: str = 'X', log_prob_collection: str = 'log_prob',
                       rng_stream: str = 'params'):
    """
    Creates a parameter from a prior transformed from a homogeneous unconstrained base measure.

    Convert a prior into a non-Bayesian parameter, that takes a single value in the model, but still has an associated
    log_prob. The parameter is registered as a `jaxns.get_parameter` with added `_param` name suffix.

    To constrain the parameter we use a Normal parameter with centre on unit cube, and scale covering the whole cube,
    as the base representation. This base representation covers the whole real line and be reliably used with SGD, etc.

    Args:
        prior: any prior
        random_init: whether to initialise the parameter randomly or at the median of the distribution.
        param_collection: the collection to register the parameter in.
        rng_stream: the name of the random number generator stream to use for sampling.

    Returns:
        A parameter representing the prior.
    """
    if prior.name is None:
        raise ValueError("Prior must have a name to be parametrised.")
    if prior.base_ndims == 0:
        warnings.warn(f"Creating a zero-sized parameter for {prior.name}. Probably unintended.")
    if init is not None:
        # tranform: X -> U -> N
        if callable(init):
            @partial(wrap_random, rng_stream=rng_stream)
            def initialiser(key, shape, dtype):
                X = jnp.asarray(init(key, shape, dtype))
                if X.shape != shape:
                    raise ValueError(f"Initialiser callable for {prior.name} returned shape {np.shape(X)}, expected {shape}.")
                if X.dtype != dtype:
                    raise ValueError(f"Initialiser callable for {prior.name} returned dtype {X.dtype}, expected {dtype}.")
                U = prior.inverse(X)
                N = quick_unit_inverse(jnp.clip(U, 1e-6, 1 - 1e-6))
                return N
        else:
            initialiser = quick_unit_inverse(jnp.clip(prior.inverse(init), 1e-6, 1 - 1e-6))
    else:
        if random_init:
            initialiser = wrap_random(sample_quick_unit_dist, rng_stream)
        else:
            # Initialises at median of distribution using zeros.
            initialiser = jnp.zeros
    N_base_param = get_parameter(
        name=prior.name,
        shape=prior.base_shape,
        dtype=prior.base_dtype,
        init=initialiser,
        collection=param_collection
    )
    # transform [-inf, inf] -> [0,1]
    U_base_param = get_parameter(
        name=prior.name,
        shape=prior.base_shape,
        dtype=prior.base_dtype,
        init=quick_unit(N_base_param),
        collection=U_collection
    )
    X_param = get_parameter(
        name=prior.name,
        collection=X_collection,
        init=prior.forward(U_base_param)
    )
    # Register the log_prob as a parameter
    _ = get_parameter(
        name=prior.name,
        collection=log_prob_collection,
        init=prior.log_prob(X_param)
    )
    return X_param


def realise_prior(prior: AbstractPrior, U_collection: str = 'U', X_collection: str = 'X',
                  log_prob_collection: str = 'log_prob', rng_stream: str = 'U'):
    """
    Convert a prior into a non-Bayesian parameter, that takes a single value in the model, but still has an associated
    log_prob. The parameter is registered as a `jaxns.get_parameter` with added `_param` name suffix.

    To constrain the parameter we use a Normal parameter with centre on unit cube, and scale covering the whole cube,
    as the base representation. This base representation covers the whole real line and be reliably used with SGD, etc.

    Args:
        prior: any prior
        U_collection: the collection to register the parameter in for U-space.
        X_collection: the collection to register the parameter in for X-space.
        rng_stream: the name of the random number generator stream to use for sampling U.

    Returns:
        A parameter representing the prior.
    """
    if prior.name is None:
        raise ValueError("Prior must have a name to be realised.")
    # Initialises at median of distribution using zeros, else unit-normal.
    initaliser = wrap_random(jax.random.uniform, rng_stream)
    if prior.base_ndims == 0:
        warnings.warn(f"Creating a zero-sized parameter for {prior.name}. Probably unintended.")
    U_base_param = get_parameter(
        name=prior.name,
        shape=prior.base_shape,
        dtype=prior.base_dtype,
        init=initaliser,
        collection=U_collection
    )
    X_param = get_parameter(
        name=prior.name,
        collection=X_collection,
        init=prior.forward(U_base_param)
    )
    # Register the log_prob as a parameter
    _ = get_parameter(
        name=prior.name,
        collection=log_prob_collection,
        init=prior.log_prob(X_param)
    )
    return X_param



#!/usr/bin/env python

from setuptools import setup


def load_requirements(file_name):
    with open(file_name, "r") as file:
        return [line.strip() for line in file if line.strip() and not line.startswith("#")]


setup(
    install_requires=load_requirements("requirements.txt"),
    extras_require={
        "tests": load_requirements("requirements-tests.txt"),
    }
)
